//		Copyright © 1998	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//		$Log: CAMUtils.cpp,v $
//		Revision 1.4  1999/01/09 15:33:39  jjk
//		Explicitly include the prefix header file (R4 x86 support).
//
//		Revision 1.3  1998/10/23 09:10:04  jjk
//		Changes for BeOS R4b1: CAM interface is now encapsulated.
//
//		Revision 1.2  1998/06/12 09:53:41  jjk
//		Encapsulate debugging messages.
//
//		Revision 1.1  1998/06/06 15:12:59  jjk
//		Initial revision
//

//		Utility functions for setting up and executing CAM requests.

#include "Prefix.h"

#include <CAM.h>
#include <KernelExport.h>
#include <module.h>

#include "CAMUtils.h"

// Global pointer to CAM module information structure.
static cam_for_driver_module_info *pCAM = (cam_for_driver_module_info *)0;

// Initialize/uninitialize the BeOS CAM interface.
status_t
camInit(void)
{
	status_t status = B_ERROR;

	if (!pCAM) {		// already initialized?
		status = get_module(B_CAM_FOR_DRIVER_MODULE_NAME, (module_info **)&pCAM);
		if (status >= B_OK && !pCAM) {
			status = B_ERROR;
		}
	}
	
	return status;
}

status_t
camExit(void)
{
	status_t status = B_ERROR;

	if (pCAM) {			// already initialized?
		status = put_module(B_CAM_FOR_DRIVER_MODULE_NAME);
		pCAM = (cam_for_driver_module_info *)0;
	}
	
	return status;
}

// Calculate CCB length for a given XPT function.  See CAM spec.
static uint16
getCCBLength(uint8 functionCode)
{
	switch (functionCode) {

	case XPT_NOOP:
	case XPT_SCAN_BUS:
	// case XPT_SCAN_LUN:	not in Be's CAM.h?
	case XPT_RESET_BUS:
	case XPT_RESET_DEV:
		return 16;

	case XPT_SDEV_TYPE:
		return 17;

	case XPT_REL_SIMQ:
	case XPT_ABORT:
	case XPT_TERM_IO:
	case XPT_NOTIFY_ACK:
		return 20;

	case XPT_GDEV_TYPE:
		return 21;

	case XPT_ENG_INQ:
		return 24;

	case XPT_SASYNC_CB:
		return 29;

	case XPT_IMMED_NOTIFY :
		return 40;

	case XPT_PATH_INQ:
		return 84;

	case XPT_SCSI_IO:
	case XPT_TARGET_IO:
	case XPT_ACCEPT_TARG :
	case XPT_CONT_TARG:			
	case XPT_ENG_EXEC:
		return sizeof(CCB_SCSIIO);

	case XPT_EN_LUN:
		return sizeof(CCB_ENABLE_LUN);
	}
	
	// This should never happen.
	ddprintf(("ST - invalid XPT function code 0x%02x.\n", functionCode));
	return 0;
}

// Allocate and initialize a CCB structure.
CCB_HEADER *
getCCB(
	uint8 functionCode, uint8 pathID, uint8 targetID, uint8 lun, uint32 flags)
{
	if (!pCAM) {
		return (CCB_HEADER *)0;
	}
	
	CCB_HEADER *const pCCB = pCAM->xpt_ccb_alloc();
	if (pCCB) {
		// Initialize header fields.
		pCCB->my_addr = pCCB;
		pCCB->cam_ccb_len = getCCBLength(functionCode);
		pCCB->cam_func_code = functionCode;
		pCCB->cam_path_id = pathID;
		pCCB->cam_target_id = targetID;
		pCCB->cam_target_lun = lun;
		pCCB->cam_flags = flags;
	}
	
	return pCCB;
}

// Release a CCB structure.
void
freeCCB(CCB_HEADER *pCCB)
{
	if (pCAM) {
		pCAM->xpt_ccb_free(pCCB);
	}
}

// Perform an XPT action using the given CCB.
long
xptAction(CCB_HEADER *pCCB)
{
	if (pCAM) {
		return pCAM->xpt_action(pCCB);
	}
	return CAM_FAILURE;
}

// Inquire the highest assigned path ID.
uint8
getHighestPathID(void)
{
	uint8 pathID = 0xff;
	
	// Issue a PATH INQUIRY request.
	CCB_PATHINQ *const pCCB = (CCB_PATHINQ *)getCCB(XPT_PATH_INQ, 0xff,0,0, 0);
	if (pCCB) {
		pCCB->cam_osd_usage = 0;
		
		long success = pCAM->xpt_action((CCB_HEADER *)pCCB);
		if (success == CAM_SUCCESS && pCCB->cam_ch.cam_status == CAM_REQ_CMP) {
			pathID = pCCB->cam_hpath_id;
		}
		
		pCAM->xpt_ccb_free(pCCB);
	}
	
	return pathID;
}

// Inquire device type information.
status_t
getDeviceType(uint8 pathID, uint8 targetID, uint8 lun,
				  uint8 &deviceType, SCSIInquiryData *const pInquiryData)
{
	status_t status;
	
	// Issue a GET DEVICE TYPE request.
	CCB_GETDEV *const pCCB =
		(CCB_GETDEV *)getCCB(XPT_GDEV_TYPE, pathID, targetID, lun, 0);
	if (pCCB) {
		pCCB->cam_inq_data = (char *)pInquiryData;
		
		if (pCAM->xpt_action((CCB_HEADER *)pCCB) == CAM_SUCCESS
			 && pCCB->cam_ch.cam_status == CAM_REQ_CMP) {
			deviceType = pCCB->cam_pd_type;
			status = B_OK;
		} else {
			status = B_ERROR;
		}
		
		pCAM->xpt_ccb_free(pCCB);

	} else {
		status = B_NO_MEMORY;
	}
	
	return status;
}

static char RCSid[] = "$Id: CAMUtils.cpp,v 1.4 1999/01/09 15:33:39 jjk Exp $";
