//		Copyright © 1998	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//		$Log: SCSICommands.cpp,v $
//		Revision 1.9  1999/01/09 15:33:39  jjk
//		Explicitly include the prefix header file (R4 x86 support).
//
//		Revision 1.8  1998/10/23 09:13:58  jjk
//		Changes for BeOS R4b1: CAM interface is now encapsulated.
//
//		Revision 1.7  1998/06/21 10:54:05  jjk
//		Report sense key and additional sense code in human readable form.
//
//		Revision 1.6  1998/06/12 09:53:41  jjk
//		Encapsulate debugging messages.
//
//		Revision 1.5  1998/06/11 16:12:32  jjk
//		Added LOAD UNLOAD command.
//
//		Revision 1.4  1998/06/11 10:21:35  jjk
//		Reduce number of retries for deviceReady().
//
//		Revision 1.3  1998/06/11 09:26:11  jjk
//		Avoid long delays; handle non-ready status elsewhere.
//
//		Revision 1.2  1998/06/07 16:27:20  jjk
//		Remove unused parameters.
//		Use longer delay when waiting for tape to become ready.
//		Use new utility functions for data conversion.
//
//		Revision 1.1  1998/06/06 15:16:03  jjk
//		Initial revision
//

//		Execution of SCSI commands.

#include "Prefix.h"

#include <CAM.h>
#include <KernelExport.h>
#include <OS.h>

#include "CAMUtils.h"
#include "SCSICommands.h"
#include "SCSIData.h"
#include "SCSIDefs.h"
#include "StDevices.h"

// Delay between retries.
const bigtime_t kRetryDelay = 100000;		// 100ms

//----------------------------------------------------------------------------
// Interpret CAM/SCSI status codes.
static status_t
decodeStatus(uint8 camStatus, uint8 scsiStatus, PhysicalDevice &device,
				 bool *&pUnitAttention)
{
	device.mSenseValid = (camStatus & CAM_AUTOSNS_VALID) != 0;
	device.mFilemark = device.mSetmark = device.mEOM = device.mEOD = false;
	device.mResidual = 0;
	
	if (pUnitAttention) {
		*pUnitAttention = false;
	}
	
	switch (camStatus & CAM_STATUS_MASK) {

	case CAM_REQ_CMP:
		// No error.
		return kStatusOK;
		
	case CAM_REQ_CMP_ERR:
		// Interpret SCSI error codes.
		switch (scsiStatus) {
		
		case SCSI_ST_CHECK_CONDITION:
			if (device.mSenseValid
				 && device.mSense.errorCode == SCSISenseData::kCurrentErrors) {
				const SCSISenseData &sense = device.mSense;
				
				// Get information field, if valid.
				if (sense.valid) {
					device.mResidual = sense.information();
				}
				
				// Check for Filemark/Setmark/EOM conditions.
				if (sense.filemark) {
					if (sense.asc == 0 && sense.asq == 3) {
						// Setmark detected.
						device.mSetmark = true;
					} else {
						device.mFilemark = true;
					}
				}
				device.mEOM = sense.eom;

				// Check sense key.
				switch (sense.senseKey) {

			 	case SCSISenseData::kNoSense:
			 	case SCSISenseData::kRecoveredError:
			 	case SCSISenseData::kMediumError:
					if (sense.filemark || sense.eom || sense.ili) {
						// May not be a fatal error - the caller has to decide.
						return kStatusNotComplete;
					}
					return B_IO_ERROR;
				
			 	case SCSISenseData::kNotReady:
			 		// Retry.
			 		return kStatusRetry;
			 	
			 	case SCSISenseData::kUnitAttention:
			 		// The first time we receive a UNIT ATTENTION condition,
			 		// we take note of the fact and retry.  All subsequent
			 		// occurrences will cause an error.
			 		if (pUnitAttention) {
			 			*pUnitAttention = true;
			 			pUnitAttention = 0;
				 		return kStatusRetry;
				 	}
				 	return B_IO_ERROR;

				case SCSISenseData::kBlankCheck:
					// Not a fatal error.
					device.mEOD = true;
					return kStatusNotComplete;
				
				case SCSISenseData::kDataProtect:
					return B_PERMISSION_DENIED;
				
				case SCSISenseData::kVolumeOverflow:
					return B_DEVICE_FULL;
				
			 	default:
					return B_IO_ERROR;
			 	}
			}
			// No sense data we can interpret, give up.
			return B_IO_ERROR;
		
		case SCSI_ST_BUSY:
		case SCSI_ST_RESERVATION_CONFLICT:
			// Retry.
			return kStatusRetry;
		}
		break;
	
	case CAM_BUSY:
	case CAM_CMD_TIMEOUT:
		// Retry.
		return kStatusRetry;
	}

	// Give up.
	return B_IO_ERROR;
}

// Report sense information for debugging.
#if ST_DEBUG
static const char *
decodeSenseKey(uint8 senseKey)
{
	static const char *kSenseKeyNames[] = {
		"NO SENSE",
		"RECOVERED ERROR",
		"NOT READY",
		"MEDIUM ERROR",
		"HARDWARE ERROR",
		"ILLEGAL REQUEST",
		"UNIT ATTENTION",
		"DATA PROTECT",
		"BLANK CHECK",
		"VENDOR-SPECIFIC",
		"COPY ABORTED",
		"ABORTED COMMAND",
		"EQUAL",
		"VOLUME OVERFLOW",
		"MISCOMPARE"
	};
	const int kNumKeys = sizeof(kSenseKeyNames) / sizeof(const char *);

	if (senseKey < kNumKeys) {
		return kSenseKeyNames[senseKey];
	}	
	return "UNKNOWN SENSE KEY";
}

static const char *
decodeAdditionalSenseCode(uint8 asc, uint8 asq)
{
	static const struct AdditionalSenseCode {
		uint16 mCode;
		const char *mpName;
	} kAdditionalSenseCodes[] = {
		{	0x0000, "NO ADDITIONAL SENSE INFORMATION"	},
		{	0x0001, "FILEMARK DETECTED"	},
		{	0x0002, "END-OF-PARTITION/MEDIUM DETECTED"	},
		{	0x0003, "SETMARK DETECTED"	},
		{	0x0004, "BEGINNING-OF-PARTITION/MEDIUM DETECTED"	},
		{	0x0005, "END-OF-DATA DETECTED"	},
		{	0x0006, "I/O PROCESS TERMINATED"	},
		{	0x0011, "AUDIO PLAY OPERATION IN PROGRESS"	},
		{	0x0012, "AUDIO PLAY OPERATION PAUSED"	},
		{	0x0013, "AUDIO PLAY OPERATION SUCCESSFULLY COMPLETED"	},
		{	0x0014, "AUDIO PLAY OPERATION STOPPED DUE TO ERROR"	},
		{	0x0015, "NO CURRENT AUDIO STATUS TO RETURN"	},
		{	0x0100, "NO INDEX/SECTOR SIGNAL"	},
		{	0x0200, "NO SEEK COMPLETE"	},
		{	0x0300, "PERIPHERAL DEVICE WRITE FAULT"	},
		{	0x0301, "NO WRITE CURRENT"	},
		{	0x0302, "EXCESSIVE WRITE ERRORS"	},
		{	0x0400, "LOGICAL UNIT NOT READY, CAUSE NOT REPORTABLE"	},
		{	0x0401, "LOGICAL UNIT IS IN PROCESS OF BECOMING READY"	},
		{	0x0402, "LOGICAL UNIT NOT READY, INITIALIZING COMMAND REQUIRED"	},
		{	0x0403, "LOGICAL UNIT NOT READY, MANUAL INTERVENTION REQUIRED"	},
		{	0x0404, "LOGICAL UNIT NOT READY, FORMAT IN PROGRESS"	},
		{	0x0500, "LOGICAL UNIT DOES NOT RESPOND TO SELECTION"	},
		{	0x0600, "NO REFERENCE POSITION FOUND"	},
		{	0x0700, "MULTIPLE PERIPHERAL DEVICES SELECTED"	},
		{	0x0800, "LOGICAL UNIT COMMUNICATION FAILURE"	},
		{	0x0801, "LOGICAL UNIT COMMUNICATION TIME-OUT"	},
		{	0x0802, "LOGICAL UNIT COMMUNICATION PARITY ERROR"	},
		{	0x0900, "TRACK FOLLOWING ERROR"	},
		{	0x0901, "TRACKING SERVO FAILURE"	},
		{	0x0902, "FOCUS SERVO FAILURE"	},
		{	0x0903, "SPINDLE SERVO FAILURE"	},
		{	0x0A00, "ERROR LOG OVERFLOW"	},
		{	0x0C00, "WRITE ERROR"	},
		{	0x0C01, "WRITE ERROR RECOVERED WITH AUTO REALLOCATION"	},
		{	0x0C02, "WRITE ERROR - AUTO REALLOCATION FAILED"	},
		{	0x1000, "ID CRC OR ECC ERROR"	},
		{	0x1100, "UNRECOVERED READ ERROR"	},
		{	0x1101, "READ RETRIES EXHAUSTED"	},
		{	0x1102, "ERROR TOO LONG TO CORRECT"	},
		{	0x1103, "MULTIPLE READ ERRORS"	},
		{	0x1104, "UNRECOVERED READ ERROR - AUTO REALLOCATE FAILED"	},
		{	0x1105, "L-EC UNCORRECTABLE ERROR"	},
		{	0x1106, "CIRC UNRECOVERED ERROR"	},
		{	0x1107, "DATA RESYNCHRONIZATION ERROR"	},
		{	0x1108, "INCOMPLETE BLOCK READ"	},
		{	0x1109, "NO GAP FOUND"	},
		{	0x110A, "MISCORRECTED ERROR"	},
		{	0x110B, "UNRECOVERED READ ERROR - RECOMMEND REASSIGNMENT"	},
		{	0x110C, "UNRECOVERED READ ERROR - RECOMMEND REWRITE THE DATA"	},
		{	0x1200, "ADDRESS MARK NOT FOUND FOR ID FIELD"	},
		{	0x1300, "ADDRESS MARK NOT FOUND FOR DATA FIELD"	},
		{	0x1400, "RECORDED ENTITY NOT FOUND"	},
		{	0x1401, "RECORD NOT FOUND"	},
		{	0x1402, "FILEMARK OR SETMARK NOT FOUND"	},
		{	0x1403, "END-OF-DATA NOT FOUND"	},
		{	0x1404, "BLOCK SEQUENCE ERROR"	},
		{	0x1500, "RANDOM POSITIONING ERROR"	},
		{	0x1501, "MECHANICAL POSITIONING ERROR"	},
		{	0x1502, "POSITIONING ERROR DETECTED BY READ OF MEDIUM"	},
		{	0x1600, "DATA SYNCHRONIZATION MARK ERROR"	},
		{	0x1700, "RECOVERED DATA WITH NO ERROR CORRECTION APPLIED"	},
		{	0x1701, "RECOVERED DATA WITH RETRIES"	},
		{	0x1702, "RECOVERED DATA WITH POSITIVE HEAD OFFSET"	},
		{	0x1703, "RECOVERED DATA WITH NEGATIVE HEAD OFFSET"	},
		{	0x1704, "RECOVERED DATA WITH RETRIES AND/OR CIRC APPLIED"	},
		{	0x1705, "RECOVERED DATA USING PREVIOUS SECTOR ID"	},
		{	0x1706, "RECOVERED DATA WITHOUT ECC - DATA AUTO-REALLOCATED"	},
		{	0x1707, "RECOVERED DATA WITHOUT ECC - RECOMMEND REASSIGNMENT"	},
		{	0x1708, "RECOVERED DATA WITHOUT ECC - RECOMMEND REWRITE"	},
		{	0x1800, "RECOVERED DATA WITH ERROR CORRECTION APPLIED"	},
		{	0x1801, "RECOVERED DATA WITH ERROR CORRECTION & RETRIES APPLIED"	},
		{	0x1802, "RECOVERED DATA - DATA AUTO-REALLOCATED"	},
		{	0x1803, "RECOVERED DATA WITH CIRC"	},
		{	0x1804, "RECOVERED DATA WITH L-EC"	},
		{	0x1805, "RECOVERED DATA - RECOMMEND REASSIGNMENT"	},
		{	0x1806, "RECOVERED DATA - RECOMMEND REWRITE"	},
		{	0x1900, "DEFECT LIST ERROR"	},
		{	0x1901, "DEFECT LIST NOT AVAILABLE"	},
		{	0x1902, "DEFECT LIST ERROR IN PRIMARY LIST"	},
		{	0x1903, "DEFECT LIST ERROR IN GROWN LIST"	},
		{	0x1A00, "PARAMETER LIST LENGTH ERROR"	},
		{	0x1B00, "SYNCHRONOUS DATA TRANSFER ERROR"	},
		{	0x1C00, "DEFECT LIST NOT FOUND"	},
		{	0x1C01, "PRIMARY DEFECT LIST NOT FOUND"	},
		{	0x1C02, "GROWN DEFECT LIST NOT FOUND"	},
		{	0x1D00, "MISCOMPARE DURING VERIFY OPERATION"	},
		{	0x1E00, "RECOVERED ID WITH ECC CORRECTION"	},
		{	0x2000, "INVALID COMMAND OPERATION CODE"	},
		{	0x2100, "LOGICAL BLOCK ADDRESS OUT OF RANGE"	},
		{	0x2101, "INVALID ELEMENT ADDRESS"	},
		{	0x2200, "ILLEGAL FUNCTION"	},
		{	0x2400, "INVALID FIELD IN CDB"	},
		{	0x2500, "LOGICAL UNIT NOT SUPPORTED"	},
		{	0x2600, "INVALID FIELD IN PARAMETER LIST"	},
		{	0x2601, "PARAMETER NOT SUPPORTED"	},
		{	0x2602, "PARAMETER VALUE INVALID"	},
		{	0x2603, "THRESHOLD PARAMETERS NOT SUPPORTED"	},
		{	0x2700, "WRITE PROTECTED"	},
		{	0x2800, "NOT READY TO READY TRANSITION, MEDIUM MAY HAVE CHANGED"	},
		{	0x2801, "IMPORT OR EXPORT ELEMENT ACCESSED"	},
		{	0x2900, "POWER ON, RESET, OR BUS DEVICE RESET OCCURRED"	},
		{	0x2A00, "PARAMETERS CHANGED"	},
		{	0x2A01, "MODE PARAMETERS CHANGED"	},
		{	0x2A02, "LOG PARAMETERS CHANGED"	},
		{	0x2B00, "COPY CANNOT EXECUTE SINCE HOST CANNOT DISCONNECT"	},
		{	0x2C00, "COMMAND SEQUENCE ERROR"	},
		{	0x2C01, "TOO MANY WINDOWS SPECIFIED"	},
		{	0x2C02, "INVALID COMBINATION OF WINDOWS SPECIFIED"	},
		{	0x2D00, "OVERWRITE ERROR ON UPDATE IN PLACE"	},
		{	0x2F00, "COMMANDS CLEARED BY ANOTHER INITIATOR"	},
		{	0x3000, "INCOMPATIBLE MEDIUM INSTALLED"	},
		{	0x3001, "CANNOT READ MEDIUM - UNKNOWN FORMAT"	},
		{	0x3002, "CANNOT READ MEDIUM - INCOMPATIBLE FORMAT"	},
		{	0x3003, "CLEANING CARTRIDGE INSTALLED"	},
		{	0x3100, "MEDIUM FORMAT CORRUPTED"	},
		{	0x3101, "FORMAT COMMAND FAILED"	},
		{	0x3200, "NO DEFECT SPARE LOCATION AVAILABLE"	},
		{	0x3201, "DEFECT LIST UPDATE FAILURE"	},
		{	0x3300, "TAPE LENGTH ERROR"	},
		{	0x3600, "RIBBON, INK, OR TONER FAILURE"	},
		{	0x3700, "ROUNDED PARAMETER"	},
		{	0x3900, "SAVING PARAMETERS NOT SUPPORTED"	},
		{	0x3A00, "MEDIUM NOT PRESENT"	},
		{	0x3B00, "SEQUENTIAL POSITIONING ERROR"	},
		{	0x3B01, "TAPE POSITION ERROR AT BEGINNING-OF-MEDIUM"	},
		{	0x3B02, "TAPE POSITION ERROR AT END-OF-MEDIUM"	},
		{	0x3B03, "TAPE OR ELECTRONIC VERTICAL FORMS UNIT NOT READY"	},
		{	0x3B04, "SLEW FAILURE"	},
		{	0x3B05, "PAPER JAM"	},
		{	0x3B06, "FAILED TO SENSE TOP-OF-FORM"	},
		{	0x3B07, "FAILED TO SENSE BOTTOM-OF-FORM"	},
		{	0x3B08, "REPOSITION ERROR"	},
		{	0x3B09, "READ PAST END OF MEDIUM"	},
		{	0x3B0A, "READ PAST BEGINNING OF MEDIUM"	},
		{	0x3B0B, "POSITION PAST END OF MEDIUM"	},
		{	0x3B0C, "POSITION PAST BEGINNING OF MEDIUM"	},
		{	0x3B0D, "MEDIUM DESTINATION ELEMENT FULL"	},
		{	0x3B0E, "MEDIUM SOURCE ELEMENT EMPTY"	},
		{	0x3D00, "INVALID BITS IN IDENTIFY MESSAGE"	},
		{	0x3E00, "LOGICAL UNIT HAS NOT SELF-CONFIGURED YET"	},
		{	0x3F00, "TARGET OPERATING CONDITIONS HAVE CHANGED"	},
		{	0x3F01, "MICROCODE HAS BEEN CHANGED"	},
		{	0x3F02, "CHANGED OPERATING DEFINITION"	},
		{	0x3F03, "INQUIRY DATA HAS CHANGED"	},
		{	0x4000, "RAM FAILURE"	},
		{	0x4100, "DATA PATH FAILURE"	},
		{	0x4200, "POWER-ON OR SELF-TEST FAILURE"	},
		{	0x4300, "MESSAGE ERROR"	},
		{	0x4400, "INTERNAL TARGET FAILURE"	},
		{	0x4500, "SELECT OR RESELECT FAILURE"	},
		{	0x4600, "UNSUCCESSFUL SOFT RESET"	},
		{	0x4700, "SCSI PARITY ERROR"	},
		{	0x4800, "INITIATOR DETECTED ERROR MESSAGE RECEIVED"	},
		{	0x4900, "INVALID MESSAGE ERROR"	},
		{	0x4A00, "COMMAND PHASE ERROR"	},
		{	0x4B00, "DATA PHASE ERROR"	},
		{	0x4C00, "LOGICAL UNIT FAILED SELF-CONFIGURATION"	},
		{	0x4E00, "OVERLAPPED COMMANDS ATTEMPTED"	},
		{	0x5000, "WRITE APPEND ERROR"	},
		{	0x5001, "WRITE APPEND POSITION ERROR"	},
		{	0x5002, "POSITION ERROR RELATED TO TIMING"	},
		{	0x5100, "ERASE FAILURE"	},
		{	0x5200, "CARTRIDGE FAULT"	},
		{	0x5300, "MEDIA LOAD OR EJECT FAILED"	},
		{	0x5301, "UNLOAD TAPE FAILURE"	},
		{	0x5302, "MEDIUM REMOVAL PREVENTED"	},
		{	0x5400, "SCSI TO HOST SYSTEM INTERFACE FAILURE"	},
		{	0x5500, "SYSTEM RESOURCE FAILURE"	},
		{	0x5700, "UNABLE TO RECOVER TABLE-OF-CONTENTS"	},
		{	0x5800, "GENERATION DOES NOT EXIST"	},
		{	0x5900, "UPDATED BLOCK READ"	},
		{	0x5A00, "OPERATOR REQUEST OR STATE CHANGE INPUT (UNSPECIFIED)"	},
		{	0x5A01, "OPERATOR MEDIUM REMOVAL REQUEST"	},
		{	0x5A02, "OPERATOR SELECTED WRITE PROTECT"	},
		{	0x5A03, "OPERATOR SELECTED WRITE PERMIT"	},
		{	0x5B00, "LOG EXCEPTION"	},
		{	0x5B01, "THRESHOLD CONDITION MET"	},
		{	0x5B02, "LOG COUNTER AT MAXIMUM"	},
		{	0x5B03, "LOG LIST CODES EXHAUSTED"	},
		{	0x5C00, "RPL STATUS CHANGE"	},
		{	0x5C01, "SPINDLES SYNCHRONIZED"	},
		{	0x5C02, "SPINDLES NOT SYNCHRONIZED"	},
		{	0x6000, "LAMP FAILURE"	},
		{	0x6100, "VIDEO ACQUISITION ERROR"	},
		{	0x6101, "UNABLE TO ACQUIRE VIDEO"	},
		{	0x6102, "OUT OF FOCUS"	},
		{	0x6200, "SCAN HEAD POSITIONING ERROR"	},
		{	0x6300, "END OF USER AREA ENCOUNTERED ON THIS TRACK"	},
		{	0x6400, "ILLEGAL MODE FOR THIS TRACK"	},
	};
	const int kNumCodes =
		sizeof(kAdditionalSenseCodes) / sizeof(struct AdditionalSenseCode);

	const uint16 code = (asc << 8) + asq;
	// Could use binary search here.
	for (int i = 0; i < kNumCodes; ++i) {
		if (kAdditionalSenseCodes[i].mCode == code) {
			return kAdditionalSenseCodes[i].mpName;
		}
	}
	return "UNKNOWN ADDITIONAL SENSE CODE";
}

static void
reportSense(PhysicalDevice &device)
{
	ddprintf((
		"     %s%s%s%s.\n"
		"     %s.\n"
		"     Sense key specific %02x %02x%02x.\n",
		decodeSenseKey(device.mSense.senseKey),
		device.mSense.filemark ? " (FM)" : "",
		device.mSense.eom ? " (EOM)" : "",
		device.mSense.ili ? " (ILI)" : "",
		decodeAdditionalSenseCode(device.mSense.asc, device.mSense.asq),
		device.mSense.senseKeySpecific[0],
		device.mSense.senseKeySpecific[1],
		device.mSense.senseKeySpecific[2]));

	if (device.mSense.valid) {
		ddprintf(("     Information %ld.\n", device.mSense.information()));
	}
}
#endif // ST_DEBUG

//----------------------------------------------------------------------------
// Perform a simple SCSI command.
status_t
scsiCommand(PhysicalDevice &device, uint32 camFlags, CDB &cdb,
				uint32 timeout, uint32 retries,
				void *transferAddress, uint32 transferLength,
				bool *pUnitAttention)
{
	status_t status;
	
	// Try to lock the memory we want to access.
	status = lock_memory(&(device.mSense), sizeof(device.mSense),
								B_DMA_IO|B_READ_DEVICE);
	if (status < B_OK) {
		return status;
	}

	const ulong memFlags = B_DMA_IO
							   | ((camFlags & CAM_DIR_IN) ? B_READ_DEVICE : 0);
	if (transferAddress && transferLength) {
		status = lock_memory(transferAddress, transferLength, memFlags);
		if (status < B_OK) {
			(void)unlock_memory(&(device.mSense), sizeof(device.mSense),
									  B_DMA_IO|B_READ_DEVICE);
			return status;
		}
	}
	
	// OK, let's fetz.
	uint32 attempt = 0;
	for (;;) {
		// Allocate and initialize the CCB.	
		CCB_SCSIIO *pCCB =
			(CCB_SCSIIO *)getCCB(XPT_SCSI_IO,
										device.mPathID, device.mTargetID, device.mLun,
										camFlags|CAM_CDB_POINTER|CAM_DIS_DISCONNECT);
		if (!pCCB) {
			status = B_NO_MEMORY;
			break;
		}
		pCCB->cam_pdrv_ptr	= 0;
		pCCB->cam_next_ccb	= 0;
		pCCB->cam_cbfcnp		= 0;
		pCCB->cam_data_ptr	= (uint8 *)transferAddress;
		pCCB->cam_dxfer_len	= transferLength;
		pCCB->cam_sense_ptr	= (uint8 *)&(device.mSense);
		pCCB->cam_sense_len	= sizeof(device.mSense);
		pCCB->cam_cdb_len		= cdb.Length();
		pCCB->cam_sglist_cnt	= 0;
		pCCB->cam_cdb_io.cam_cdb_ptr = cdb.Address();
		pCCB->cam_timeout		= timeout;
		
		// Perform the command and retain results.
		long success	  = xptAction((CCB_HEADER *)pCCB);
		uint8 camStatus  = pCCB->cam_ch.cam_status;
		uint8 scsiStatus = pCCB->cam_scsi_status;
	
		freeCCB((CCB_HEADER *)pCCB);
		
		// Interpret the results.
		status = decodeStatus(camStatus, scsiStatus, device, pUnitAttention);
		if (success != CAM_SUCCESS
			 || (camStatus & CAM_STATUS_MASK) != CAM_REQ_CMP
			 || scsiStatus != SCSI_ST_GOOD) {
#if ST_DEBUG
			ddprintf(("ST - Request %02x %s, CAM %02x, SCSI %02x.\n",
						 *(cdb.Address()),
						 success == CAM_SUCCESS ? "success" : "failure",
						 camStatus, scsiStatus));
			reportSense(device);
#endif
		}

		// Get out if everything is OK, or if we cannot retry.
		if (status != kStatusRetry || ++attempt > retries) {
			break;
		}

		snooze(kRetryDelay);
	}

	// Unlock whatever memory we locked.
	if (transferAddress && transferLength) {
		(void)unlock_memory(transferAddress, transferLength, memFlags);
	}
	(void)unlock_memory(&(device.mSense), sizeof(device.mSense),
							  B_DMA_IO|B_READ_DEVICE);
	
	// Wow.
	return status;
}

//----------------------------------------------------------------------------
// Some commonly-used SCSI commands.

// Check if the device is ready.
status_t
deviceReady(PhysicalDevice &device, bool &unitAttention)
{
	// Send TEST UNIT READY command.
	CDBTestUnitReady cdb(device.mLun);
	return scsiCommand(device, CAM_DIR_NONE, cdb, 1, 10, 0, 0, &unitAttention);
}

// Perform a MODE SELECT.
status_t
scsiModeSelect(PhysicalDevice &device, void *pData, uint8 len)
{
	CDBModeSelect0 cdb(device.mLun, len);
	return scsiCommand(device, CAM_DIR_OUT, cdb, pData, len, 5, kRetries);
}

// Perform a MODE SENSE.
status_t
scsiModeSense(PhysicalDevice &device, uint8 pageCode, void *pData, uint8 len,
			 bool dbd, uint8 pc)
{
	CDBModeSense0 cdb(device.mLun, pageCode, len, dbd, pc);
	return scsiCommand(device, CAM_DIR_IN, cdb, pData, len, 5, kRetries);
}

// Get resp. set the current block size and density.
status_t
getBlockSize(PhysicalDevice &device, uint32 &blockSize, uint8 &density)
{
	struct {
		SCSIModeParameterHeader6 h;
		SCSIBlockDescriptor bd;
	} modeParList;
	
	status_t status =
		scsiModeSense(device, 0, &modeParList, sizeof(modeParList));
	if (status == B_OK) {
		if (modeParList.h.blockDescriptorLength != sizeof(SCSIBlockDescriptor)) {
			// This shouldn't happen - at least not with my drive :-)
			status = B_IO_ERROR;
		} else {
			blockSize = modeParList.bd.blockLength();
			density = modeParList.bd.densityCode;
		}
	}
	
	return status;
}

status_t
setBlockSize(PhysicalDevice &device, uint32 blockSize, uint8 density)
{
	struct {
		SCSIModeParameterHeader6 h;
		SCSIBlockDescriptor bd;
	} modeParList;
	
	modeParList.h.modeDataLength = 0;		// reserved for MODE SELECT
	modeParList.h.mediumType = 0;
	modeParList.h.tape.wp = 0;
	modeParList.h.tape.bufferedMode = SCSIModeParameterHeader6::kBuffered;
	modeParList.h.tape.speed = 0;
	modeParList.h.blockDescriptorLength = sizeof(SCSIBlockDescriptor);
	
	modeParList.bd.densityCode = density;
	SCSIData::ui32_3b(0, modeParList.bd.numBlocks_);
	modeParList.bd.res4 = 0;
	SCSIData::ui32_3b(blockSize, modeParList.bd.blockLength_);;

	return scsiModeSelect(device, &modeParList, sizeof(modeParList));
}

// Perform various common commands.
status_t
scsiLoadUnload(PhysicalDevice &device, bool load, bool flag, bool immediate)
{
	CDBLoadUnload cdb(device.mLun, load, flag, immediate);
	return scsiCommand(device, CAM_DIR_NONE, cdb,
							 immediate ? 5 : 900, kRetries);
}

status_t
scsiPreventMediumRemoval(PhysicalDevice &device, bool prevent)
{
	CDBMediumRemoval cdb(device.mLun, prevent);
	return scsiCommand(device, CAM_DIR_NONE, cdb, 1);
}

status_t
scsiRewind(PhysicalDevice &device, bool immediate)
{
	CDBRewind cdb(device.mLun, immediate);
	return scsiCommand(device, CAM_DIR_NONE, cdb, immediate ? 5 : 300);
}

status_t
scsiSpace(PhysicalDevice &device, uint8 code, int32 count)
{
	CDBSpace cdb(device.mLun, code, count);
	return scsiCommand(device, CAM_DIR_NONE, cdb, 3600);	// wait up to 1 hour
}

status_t
scsiWriteFilemarks(PhysicalDevice &device, uint32 count, bool writeSetmarks)
{
	CDBWriteFilemarks cdb(device.mLun, count, writeSetmarks);
	return scsiCommand(device, CAM_DIR_NONE, cdb, 10);
}

static char RCSid[] = "$Id: SCSICommands.cpp,v 1.9 1999/01/09 15:33:39 jjk Exp $";
