//		Copyright © 1998	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//		$Log: StDevices.h,v $
//		Revision 1.4  1998/10/23 09:20:56  jjk
//		Add support for multiple partitions per physical device.
//
//		Revision 1.3  1998/06/21 10:03:27  jjk
//		Backspace over second EOF at EOD when doing MTIOCTOP MTEOD.
//
//		Revision 1.2  1998/06/11 22:17:10  jjk
//		Store entire device information persistently.
//
//		Revision 1.1  1998/06/06 15:17:57  jjk
//		Initial revision
//

//		Management of device names and descriptors.

#ifndef ST_DEVICES_H
#define ST_DEVICES_H

#include <OS.h>

#include "SCSIData.h"

// Flags used during resource discovery.
enum {
	kDevicePresent = 1,
	kRemovableMedium = 2
};

//------------------------------------------------------------------------------
// Description of a physical device; used for all SCSI transactions.
class PhysicalDevice
{
public:
	// Device usage flags.  Must be set/reset atomically.
	enum {
		kInUse		= 1,
		kOpen			= 2
	};
	int32 mFlags;

	// Semaphore for locking against concurrent access.
	const sem_id mLock;
	
	// Connect ID.
	const uint8 mPathID;
	const uint8 mTargetID;
	const uint8 mLun;

	// Device properties.
	const bool mRemovableMedium;		// medium is removable

	uint32 mMinBlockSize;				// block size limits
	uint32 mMaxBlockSize;	
	uint32 mBlockSize;					// current block size (0 = variable)
	uint8 mDensity;						// density (mostly ignored)
	
	bool mWriteProtected	:1;			// medium is write protected
	bool mDataCompression:1;			// data compression is in effect
	bool mReadOK			:1;			// device is open for reading
	bool mWriteOK			:1;			//              - for writing
	bool mWrittenTo		:1;			// device has been written to
	bool mWriteEOF			:1;			// automatic writing of EOF is allowed
	bool mWrite2EOF		:1;			// device needs two EOF marks at EOD
	bool mWrittenPastEOM	:1;			// last write encountered EOM
	bool mWriteBarrier	:1;			// writes at EOM will report error
	bool mStatusPending	:1;			// read status has not yet been reported
	bool mOnline			:1;			// device is online
	
	// Raw and decoded REQUEST SENSE results of last command executed.
	SCSISenseData mSense;
	bool mSenseValid		:1;
	
	bool mFilemark			:1;			// filemark encountered
	bool mSetmark			:1;			// setmark encountered
	bool mEOD				:1;			// end-of-data encountered
	bool mEOM				:1;			// end-of-medium (BOT/EW)
	bool mBOT				:1;			// end-of-medium after backward movement
	bool mEOT				:1;			// end-of-medium after forward movement
	int32 mResidual;						// residual count

	// Constructor.
	PhysicalDevice(sem_id lock,
						uint8 pathID, uint8 targetID, uint8 lun,
						bool removableMedium)
	:	mFlags(0), mLock(lock),
		mPathID(pathID), mTargetID(targetID), mLun(lun),
		mRemovableMedium(removableMedium),
		mMinBlockSize(0), mMaxBlockSize(0), mBlockSize(0), mDensity(0),
		mSenseValid(false)
	{
		// empty
	}

	// Destructor.
	~PhysicalDevice(void)
	{
		delete_sem(mLock);
	}
	
private:
	PhysicalDevice(const PhysicalDevice &);
	PhysicalDevice &operator =(const PhysicalDevice &);
};

//------------------------------------------------------------------------------
// Auxiliary class to perform automatic locking of a physical device.
class DeviceLock
{
public:
	DeviceLock(PhysicalDevice &device)
	:	mLock(device.mLock)
	{
		while (acquire_sem(mLock) == B_INTERRUPTED) {
			// loop
		}
	}
	
	~DeviceLock(void)
	{
		release_sem(mLock);
	}

private:
	DeviceLock(const DeviceLock &);
	DeviceLock &operator =(const DeviceLock &);
	void *operator new(size_t);
	void operator delete(void *)
	{
		// never called, but needed to keep the linker happy
	}
	
	sem_id mLock;
};

//------------------------------------------------------------------------------
// Description of a logical device visible to the user (/dev/tape/scsi/...).	
class LogicalDevice
{
public:
	// Corresponding physical device.
	PhysicalDevice &mPhys;

	// Logical partition.
	const int mPartition;
	
	// Flags governing the driver's behavior.
	const bool mCompression;	// data compression enabled
	const bool mAutoRewind;		// auto rewind on close
	const bool mBSDBehavior;	// Berkeley style behavior
	
	// Constructor.
	LogicalDevice(PhysicalDevice &phys, int partition,
					  bool compression, bool bsdBehavior, bool autoRewind)
	:	mPhys(phys),
		mPartition(partition),
		mCompression(compression),
		mAutoRewind(autoRewind),
		mBSDBehavior(bsdBehavior)
	{
		// empty
	}

private:
	LogicalDevice(const LogicalDevice &);
	LogicalDevice &operator =(const LogicalDevice &);
};

//------------------------------------------------------------------------------
// Given a map of available devices, build descriptors and assign names.
status_t buildDeviceTables(const int nDevices, const uint8 *const pMap);

// Free the device tables.
void freeDeviceTables(void);

// Find device descriptor for a given name.
LogicalDevice *findDevice(const char *pName);

#endif // ST_DEVICES_H
