//		Copyright © 1998–1999  Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//		$Log: StIOControl.cpp,v $
//		Revision 1.8  1999/04/18 10:16:27  jjk
//		GCC support; add MTIOCPOS and MTIOCSETPOS.
//
//		Revision 1.7  1999/01/09 15:33:39  jjk
//		Explicitly include the prefix header file (R4 x86 support).
//
//		Revision 1.6  1998/06/21 13:22:32  jjk
//		Ensure correct status after writing filemarks/setmarks.
//
//		Revision 1.5  1998/06/21 10:03:27  jjk
//		Backspace over second EOF at EOD when doing MTIOCTOP MTEOD.
//
//		Revision 1.4  1998/06/12 09:53:41  jjk
//		Encapsulate debugging messages.
//
//		Revision 1.3  1998/06/11 16:12:32  jjk
//		Make scsiLoadUnload a public function.
//
//		Revision 1.2  1998/06/11 10:16:25  jjk
//		Try to bring drive on-line if necessary.
//
//		Revision 1.1  1998/06/06 15:19:23  jjk
//		Initial revision
//

//		Implementation of driver operation control().

#include "Prefix.h"

#include <string.h>

#include <CAM.h>
#include <Drivers.h>
#include <KernelExport.h>

#include "mtio.h"
#include "SCSICommands.h"
#include "SCSIData.h"
#include "StDevices.h"
#include "StOpenClose.h"
#include "StIOControl.h"

//------------------------------------------------------------------------------
// SCSI command execution.
static status_t
scsiErase(PhysicalDevice &device, bool long_ = true, bool immediate = false)
{
	CDBErase cdb(device.mLun, long_, immediate);
	return scsiCommand(device, CAM_DIR_NONE, cdb, immediate?5:1200, kRetries);
}

//------------------------------------------------------------------------------
// Get device icon.
static status_t
mtioGetIcon(const device_icon &request)
{
	ddprintf(("ST - Icon requested, size %ld.\n", request.icon_size));
	// ### NOT IMPLEMENTED YET
	return B_BAD_VALUE;
}

//------------------------------------------------------------------------------
// Inquire current status.
static status_t
mtioGetStatus(PhysicalDevice &device, mtget &request)
{
	memset(&request, 0, sizeof(request));
	
	// Get residual count.
	request.mt_resid = device.mResidual;
	
	// Get flags & density.
	request.mt_gstat =
		 (device.mBOT					? kFlagBOT	: 0)
		|(device.mEOD					? kFlagEOD	: 0)
		|(device.mEOT					? kFlagEOT	: 0)
		|(device.mFilemark			? kFlagEOF	: 0)
		|(device.mSetmark				? kFlagSM	: 0)
		|(device.mWriteProtected	? kFlagWP	: 0)
		|(device.mDataCompression	? kFlagCMP	: 0)
		|(device.mOnline				? kFlagONL	: 0)
		|(uint32)(device.mDensity);

	// Get block size.
	request.mt_bsize = device.mBlockSize;
	
	// If we've reached EOT, allow writing past it (YAFIYGI).
	if (device.mWrittenPastEOM) {
		device.mWriteBarrier = false;
	}
	
	return B_OK;
}

//------------------------------------------------------------------------------
// Perform tape control operations.
static status_t
mtioErase(PhysicalDevice &device)
{
	// Reject if device is not open for writing.
	if (!device.mWriteOK) {
		return B_PERMISSION_DENIED;
	}

	// Perform a long erase.  According to the standard, a device may require
	// that it is positioned at BOT, so we rewind the medium first.
	status_t status = scsiRewind(device);
	if (status == B_OK) {
		// Disable the automatic writing of filemarks.
		device.mWriteEOF = false;
		
		status = scsiErase(device);
	}
	return status;
}

static status_t
mtioOffline(PhysicalDevice &device)
{
	// Allow medium removal (if applicable).
	if (device.mRemovableMedium) {
		(void)scsiPreventMediumRemoval(device, false);
	}
	
	// Unload the medium.
	status_t status = scsiLoadUnload(device, false);
	
	if (status < B_OK) {
		// This is probably futile, but we try anyway.
		if (device.mRemovableMedium) {
			(void)scsiPreventMediumRemoval(device, true);
		}
	} else {
		// Disable the automatic writing of filemarks.
		device.mWriteEOF = false;
		
		device.mOnline = false;
	}
	return status;
}

static status_t
mtioRetension(PhysicalDevice &device)
{
	status_t status = scsiLoadUnload(device, true, true);
	if (status == B_OK) {
		// Disable the automatic writing of filemarks.
		device.mWriteEOF = false;
	}
	return status;
}

static status_t
mtioRewind(PhysicalDevice &device)
{
	status_t status = scsiRewind(device);
	if (status == B_OK) {
		// Disable the automatic writing of filemarks.
		device.mWriteEOF = false;
	}
	return status;
}

static status_t
mtioSetBlockSize(PhysicalDevice &device, uint32 blockSize)
{
	const uint32 minSize = device.mMinBlockSize;
	const uint32 maxSize = device.mMaxBlockSize;
		
	status_t status = B_OK;
	if (blockSize == 0) {
		// Variable sized blocks requested...
		if (minSize == maxSize) {
			// ...but not supported.
			status = B_BAD_VALUE;
		}

	} else if (blockSize < minSize || blockSize > maxSize) {
		status = B_BAD_VALUE;
	}
	
	if (status == B_OK) {
		// Try to set the block size.
		(void)setBlockSize(device, blockSize);
		status = getBlockSize(device, device.mBlockSize, device.mDensity);
		if (status >= B_OK && device.mBlockSize != blockSize) {
			// Something went wrong.
			status = B_BAD_VALUE;
		}
	}
	
	return status;
}

static status_t
mtioSpace(PhysicalDevice &device, uint8 code, int32 count = 1)
{
	// Do the first installment.
	status_t status = scsiSpace(device, code, count);

	// If repositioning has (at least partly) succeeded,
	// disable the automatic writing of filemarks.
	if (status == B_OK || status == kStatusNotComplete) {
		device.mWriteEOF = false;
	}
	
	// Repeat until we reach count zero or an error occurs.
	while (status == kStatusNotComplete) {
		if (device.mEOD) {
			status = B_IO_ERROR;

		} else if (device.mEOM) {
			device.mBOT = (count < 0);
			device.mEOT = (count > 0);
			status = B_IO_ERROR;
		
		} else if (device.mResidual != 0 && (device.mFilemark||device.mSetmark)) {
			// A hierarchically superior mark was encountered.
			// This is OK; it will be included in the total count.
			const int32 residual = (count < 0)
										? device.mResidual + 1
										: device.mResidual - 1;
			if (residual) {
				status = scsiSpace(device, code, residual);
			} else {
				status = B_OK;
			}
		} else {
			// This shouldn't happen.
			ddprintf(("ST - Unexpected result in mtioSpace().\n"));
			status = B_IO_ERROR;
		}
	}

	if (status == B_OK) {
		// Ensure that the device status is up-to-date.
		if (code == CDBSpace::kFilemarks && !device.mSetmark) {
			device.mFilemark = true;
		} else if (code == CDBSpace::kSetmarks) {
			device.mSetmark = true;
		}
	}
	
	return status;
}

static status_t
mtioWriteFilemarks(PhysicalDevice &device, uint32 count, bool writeSetmarks)
{
	// Reject if device is not open for writing.
	if (!device.mWriteOK) {
		return B_PERMISSION_DENIED;
	}

	status_t status = scsiWriteFilemarks(device, count, writeSetmarks);
	if (status == B_OK
		 || (status == kStatusNotComplete && (uint32)device.mResidual < count)) {
		// Ensure that the device status is up-to-date.
		device.mFilemark = !(device.mSetmark = writeSetmarks);
	}
	return (status == kStatusNotComplete) ? B_OK : status;
}

static status_t
mtioTapeOperation(LogicalDevice &logicalDevice, mtop &request)
{
	PhysicalDevice &device = logicalDevice.mPhys;

	// Reject if the device is offline.
	if (!device.mOnline) {
		status_t status = mountDevice(logicalDevice);
		if (status < B_OK) {
			return status;
		}
	}

	// Clear any pending read status.
	device.mStatusPending = false;

	// Perform specific operation.
	switch (request.mt_op) {

	// Write a number of filemarks resp. setmarks.
	case MTWEOF:
		if (request.mt_count < 0) {
			return B_BAD_VALUE;
		}
		return mtioWriteFilemarks(device, request.mt_count, false);
	
	case MTWSS:
		if (request.mt_count < 0) {
			return B_BAD_VALUE;
		}
		return mtioWriteFilemarks(device, request.mt_count, true);

	// Space across records, filemarks or setmarks.
	case MTBSR:
		request.mt_count = - request.mt_count;
	case MTFSR:
		return mtioSpace(device, CDBSpace::kBlocks, request.mt_count);

	case MTBSF:
		request.mt_count = - request.mt_count;
	case MTFSF:
		return mtioSpace(device, CDBSpace::kFilemarks, request.mt_count);

	case MTBSS:
		request.mt_count = - request.mt_count;
	case MTFSS:
		return mtioSpace(device, CDBSpace::kSetmarks, request.mt_count);

	// Space to end of data.
	case MTEOD:
		{	status_t status = mtioSpace(device, CDBSpace::kEndOfData);
			if (status == B_OK && device.mWrite2EOF) {
				// Move backwards over the second filemark that should be here.
				status = mtioSpace(device, CDBSpace::kFilemarks, -1);
			}
			return status;
		}
	
	// Rewind the medium.
	case MTREW:
		return mtioRewind(device);

	// Rewind and go offline.
	case MTOFFL:
		return mtioOffline(device);

	// Retension the medium.
	case MTRETENS:
		return mtioRetension(device);

	// Erase the medium.
	case MTERASE:
		return mtioErase(device);

	// Set block size.
	case MTSETBSIZ:
		if (request.mt_count < 0) {
			return B_BAD_VALUE;
		}
		return mtioSetBlockSize(device, (uint32)(request.mt_count));

	// Flush write buffers out to medium.
	case MTNOP:
		if (device.mWriteOK) {
			return scsiWriteFilemarks(device, 0);
		}
		return B_OK;
	
	// Unsupported operations.
	case MTCACHE:
	case MTNOCACHE:
		return B_OK;
	}

	// Operation not supported.
	return B_BAD_VALUE;
}

//------------------------------------------------------------------------------
// Inquire resp. change tape position.
static status_t
mtioGetPosition(PhysicalDevice &device, mtpos &position)
{
	CDBReadPosition cdb(device.mLun);
	SCSIReadPositionData data;
	status_t status = scsiCommand(device, CAM_DIR_IN, cdb, &data, sizeof(data),
											5, kRetries);
	if (status == B_OK) {
		if (data.bpu) {
			// Block position unknown.
			status = B_IO_ERROR;
		} else {
			position.mt_blkno = data.firstBlockLocation();
		}
	}
	
	return status;
}

static status_t
mtioSetPosition(PhysicalDevice &device, mtpos &position)
{
	CDBLocate cdb(device.mLun, position.mt_blkno);
	status_t status = scsiCommand(device, CAM_DIR_NONE, cdb, 300, kRetries);

	// If repositioning succeeded, disable the automatic writing of filemarks.
	if (status == B_OK) {
		device.mWriteEOF = false;
	}
	
	return status;
}

//------------------------------------------------------------------------------
// Perform ioctl() operations on a tape device.
status_t
stControl(void *pCookie, uint32 op, void *pData, size_t /*len*/)
{
	LogicalDevice *pLogicalDevice = (LogicalDevice *)pCookie;
	PhysicalDevice &device = pLogicalDevice->mPhys;
	DeviceLock lock(device);
	
	// Reject if device has been closed.
	if (!(device.mFlags & PhysicalDevice::kOpen)) {
		return B_FILE_ERROR;
	}

	// Dispatch according to request.
	switch (op) {

	// Operations defined in mtio.h.
	case MTIOCGET:
		return mtioGetStatus(device, *(mtget *)pData);
	
	case MTIOCTOP:
		return mtioTapeOperation(*pLogicalDevice, *(mtop *)pData);
	
	case MTIOCPOS:
		return mtioGetPosition(device, *(mtpos *)pData);
	
	case MTIOCSETPOS:
		return mtioSetPosition(device, *(mtpos *)pData);
	
	// Unsupported operations.  I don't even know what they're *supposed* to do.
	case MTIOCIEOT:
	case MTIOCEEOT:
		break;
	
	// Standard Be operation codes that we support.
	case B_EJECT_DEVICE:
		{	mtop op = { MTOFFL, 0 };
			return mtioTapeOperation(*pLogicalDevice, op);
		}
	
	case B_GET_ICON:
		return mtioGetIcon(*(const device_icon *)pData);
	}
	
	return B_BAD_VALUE;
}

static char RCSid[] = "$Id: StIOControl.cpp,v 1.8 1999/04/18 10:16:27 jjk Exp $";
